@@ -1,13 +1,14 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE FlexibleContexts  #-}
 {-# LANGUAGE CPP #-}
-
-#ifdef GENERICS
 {-# LANGUAGE DefaultSignatures
            , TypeOperators
            , BangPatterns
            , KindSignatures
            , ScopedTypeVariables #-}
+
+#ifndef MIN_VERSION_base
+#define MIN_VERSION_base(x,y,z) 1
 #endif
 
 -----------------------------------------------------------------------------
@@ -37,6 +38,10 @@ module Data.Serialize (
     , module Data.Serialize.Get
     , module Data.Serialize.Put
     , module Data.Serialize.IEEE754
+
+    -- * Generic deriving
+    , GSerializePut(..)
+    , GSerializeGet(..)
     ) where
 
 import Data.Serialize.Put
@@ -63,11 +68,16 @@ import qualified Data.Ratio           as R
 import qualified Data.Tree            as T
 import qualified Data.Sequence        as Seq
 
-#ifdef GENERICS
 import GHC.Generics
+
+#if !(MIN_VERSION_base(4,8,0))
 import Control.Applicative ((*>),(<*>),(<$>),pure)
 #endif
 
+#if MIN_VERSION_base(4,8,0)
+import Numeric.Natural
+#endif
+
 ------------------------------------------------------------------------
 
 
@@ -84,13 +94,11 @@ class Serialize t where
     -- | Decode a value in the Get monad
     get :: Get t
 
-#ifdef GENERICS
-    default put :: (Generic t, GSerialize (Rep t)) => Putter t
+    default put :: (Generic t, GSerializePut (Rep t)) => Putter t
     put = gPut . from
 
-    default get :: (Generic t, GSerialize (Rep t)) => Get t
+    default get :: (Generic t, GSerializeGet (Rep t)) => Get t
     get = to <$> gGet
-#endif
 
 ------------------------------------------------------------------------
 -- Wrappers to run the underlying monad
@@ -132,15 +140,39 @@ instance Serialize () where
     put ()  = return ()
     get     = return ()
 
+{-# INLINE boolToWord8 #-}
+boolToWord8 :: Bool -> Word8
+boolToWord8 False = 0
+boolToWord8 True = 1
+
+{-# INLINE boolFromWord8 #-}
+boolFromWord8 :: Word8 -> Get Bool
+boolFromWord8 0 = return False
+boolFromWord8 1 = return True
+boolFromWord8 w = fail ("Invalid Bool encoding " ++ show w)
+
+{-# INLINE orderingToWord8 #-}
+orderingToWord8 :: Ordering -> Word8
+orderingToWord8 LT = 0
+orderingToWord8 EQ = 1
+orderingToWord8 GT = 2
+
+{-# INLINE orderingFromWord8 #-}
+orderingFromWord8 :: Word8 -> Get Ordering
+orderingFromWord8 0 = return LT
+orderingFromWord8 1 = return EQ
+orderingFromWord8 2 = return GT
+orderingFromWord8 w = fail ("Invalid Ordering encoding " ++ show w)
+
 -- Bools are encoded as a byte in the range 0 .. 1
 instance Serialize Bool where
-    put     = putWord8 . fromIntegral . fromEnum
-    get     = liftM (toEnum . fromIntegral) getWord8
+    put     = putWord8 . boolToWord8
+    get     = boolFromWord8 =<< getWord8
 
 -- Values of type 'Ordering' are encoded as a byte in the range 0 .. 2
 instance Serialize Ordering where
-    put     = putWord8 . fromIntegral . fromEnum
-    get     = liftM (toEnum . fromIntegral) getWord8
+    put     = putWord8 . orderingToWord8
+    get     = orderingFromWord8 =<< getWord8
 
 ------------------------------------------------------------------------
 -- Words and Ints
@@ -238,13 +270,13 @@ instance Serialize Integer where
 --
 -- Fold and unfold an Integer to and from a list of its bytes
 --
-unroll :: Integer -> [Word8]
+unroll :: (Integral a, Bits a) => a -> [Word8]
 unroll = unfoldr step
   where
     step 0 = Nothing
     step i = Just (fromIntegral i, i `shiftR` 8)
 
-roll :: [Word8] -> Integer
+roll :: (Integral a, Bits a) => [Word8] -> a
 roll   = foldr unstep 0
   where
     unstep b a = a `shiftL` 8 .|. fromIntegral b
@@ -253,6 +285,31 @@ instance (Serialize a,Integral a) => Serialize (R.Ratio a) where
     put r = put (R.numerator r) >> put (R.denominator r)
     get = liftM2 (R.%) get get
 
+#if MIN_VERSION_base(4,8,0)
+-- Fixed-size type for a subset of Natural
+type NaturalWord = Word64
+
+instance Serialize Natural where
+    {-# INLINE put #-}
+    put n | n <= hi = do
+        putWord8 0
+        put (fromIntegral n :: NaturalWord)  -- fast path
+     where
+        hi = fromIntegral (maxBound :: NaturalWord) :: Natural
+
+    put n = do
+        putWord8 1
+        put (unroll (abs n))         -- unroll the bytes
+
+    {-# INLINE get #-}
+    get = do
+        tag <- get :: Get Word8
+        case tag of
+            0 -> liftM fromIntegral (get :: Get NaturalWord)
+            _ -> do bytes <- get
+                    return $! roll bytes
+#endif
+
 ------------------------------------------------------------------------
 
 -- Safely wrap `chr` to avoid exceptions. 
@@ -458,12 +515,12 @@ instance (Serialize e) => Serialize (Seq.Seq e) where
 -- Floating point
 
 instance Serialize Double where
-    put d = put (decodeFloat d)
-    get   = liftM2 encodeFloat get get
+    put = putFloat64be
+    get = getFloat64be
 
 instance Serialize Float where
-    put f = put (decodeFloat f)
-    get   = liftM2 encodeFloat get get
+    put = putFloat32be
+    get = getFloat32be
 
 ------------------------------------------------------------------------
 -- Trees
@@ -487,39 +544,48 @@ instance (Serialize i, Ix i, Serialize e, IArray UArray e)
     put = putIArrayOf put put
     get = getIArrayOf get get
 
-#ifdef GENERICS
 ------------------------------------------------------------------------
 -- Generic Serialze
 
-class GSerialize f where
+class GSerializePut f where
     gPut :: Putter (f a)
+
+class GSerializeGet f where
     gGet :: Get (f a)
 
-instance GSerialize a => GSerialize (M1 i c a) where
+instance GSerializePut a => GSerializePut (M1 i c a) where
     gPut = gPut . unM1
+    {-# INLINE gPut #-}
+
+instance GSerializeGet a => GSerializeGet (M1 i c a) where
     gGet = M1 <$> gGet
-    {-# INLINE gPut #-}
     {-# INLINE gGet #-}
 
-instance Serialize a => GSerialize (K1 i a) where
+instance Serialize a => GSerializePut (K1 i a) where
     gPut = put . unK1
+    {-# INLINE gPut #-}
+
+instance Serialize a => GSerializeGet (K1 i a) where
     gGet = K1 <$> get
-    {-# INLINE gPut #-}
     {-# INLINE gGet #-}
 
-instance GSerialize U1 where
+instance GSerializePut U1 where
     gPut _ = pure ()
+    {-# INLINE gPut #-}
+
+instance GSerializeGet U1 where
     gGet   = pure U1
-    {-# INLINE gPut #-}
     {-# INLINE gGet #-}
 
-instance (GSerialize a, GSerialize b) => GSerialize (a :*: b) where
+instance (GSerializePut a, GSerializePut b) => GSerializePut (a :*: b) where
     gPut (a :*: b) = gPut a *> gPut b
-    gGet = (:*:) <$> gGet  <*> gGet
     {-# INLINE gPut #-}
+
+instance (GSerializeGet a, GSerializeGet b) => GSerializeGet (a :*: b) where
+    gGet = (:*:) <$> gGet  <*> gGet
     {-# INLINE gGet #-}
 
--- The following GSerialize instance for sums has support for serializing types
+-- The following GSerialize* instance for sums has support for serializing types
 -- with up to 2^64-1 constructors. It will use the minimal number of bytes
 -- needed to encode the constructor. For example when a type has 2^8
 -- constructors or less it will use a single byte to encode the constructor. If
@@ -529,20 +595,20 @@ instance (GSerialize a, GSerialize b) => GSerialize (a :*: b) where
 #define PUTSUM(WORD) GUARD(WORD) = putSum (0 :: WORD) (fromIntegral size)
 #define GETSUM(WORD) GUARD(WORD) = (get :: Get WORD) >>= checkGetSum (fromIntegral size)
 
-instance ( PutSum     a, PutSum     b
-         , GetSum     a, GetSum     b
-         , GSerialize a, GSerialize b
-         , SumSize    a, SumSize    b) => GSerialize (a :+: b) where
+instance ( PutSum        a, PutSum        b
+         , SumSize       a, SumSize       b) => GSerializePut (a :+: b) where
     gPut | PUTSUM(Word8) | PUTSUM(Word16) | PUTSUM(Word32) | PUTSUM(Word64)
          | otherwise = sizeError "encode" size
       where
         size = unTagged (sumSize :: Tagged (a :+: b) Word64)
+    {-# INLINE gPut #-}
 
+instance ( GetSum        a, GetSum        b
+         , SumSize       a, SumSize       b) => GSerializeGet (a :+: b) where
     gGet | GETSUM(Word8) | GETSUM(Word16) | GETSUM(Word32) | GETSUM(Word64)
          | otherwise = sizeError "decode" size
       where
         size = unTagged (sumSize :: Tagged (a :+: b) Word64)
-    {-# INLINE gPut #-}
     {-# INLINE gGet #-}
 
 sizeError :: Show size => String -> size -> error
@@ -553,7 +619,7 @@ sizeError s size = error $ "Can't " ++ s ++ " a type with " ++ show size ++ " co
 class PutSum f where
     putSum :: (Num word, Bits word, Serialize word) => word -> word -> Putter (f a)
 
-instance (PutSum a, PutSum b, GSerialize a, GSerialize b) => PutSum (a :+: b) where
+instance (PutSum a, PutSum b) => PutSum (a :+: b) where
     putSum !code !size s = case s of
                              L1 x -> putSum code           sizeL x
                              R1 x -> putSum (code + sizeL) sizeR x
@@ -566,7 +632,7 @@ instance (PutSum a, PutSum b, GSerialize a, GSerialize b) => PutSum (a :+: b) wh
           sizeR = size - sizeL
     {-# INLINE putSum #-}
 
-instance GSerialize a => PutSum (C1 c a) where
+instance GSerializePut a => PutSum (C1 c a) where
     putSum !code _ x = put code *> gPut x
     {-# INLINE putSum #-}
 
@@ -581,7 +647,7 @@ checkGetSum size code | code < size = getSum code size
 class GetSum f where
     getSum :: (Ord word, Num word, Bits word) => word -> word -> Get (f a)
 
-instance (GetSum a, GetSum b, GSerialize a, GSerialize b) => GetSum (a :+: b) where
+instance (GetSum a, GetSum b) => GetSum (a :+: b) where
     getSum !code !size | code < sizeL = L1 <$> getSum code           sizeL
                        | otherwise    = R1 <$> getSum (code - sizeL) sizeR
         where
@@ -593,7 +659,7 @@ instance (GetSum a, GetSum b, GSerialize a, GSerialize b) => GetSum (a :+: b) wh
           sizeR = size - sizeL
     {-# INLINE getSum #-}
 
-instance GSerialize a => GetSum (C1 c a) where
+instance GSerializeGet a => GetSum (C1 c a) where
     getSum _ _ = gGet
     {-# INLINE getSum #-}
 
@@ -610,4 +676,3 @@ instance (SumSize a, SumSize b) => SumSize (a :+: b) where
 
 instance SumSize (C1 c a) where
     sumSize = Tagged 1
-#endif
